import importlib
import importlib.resources
import os
import re
import socket
import time
from datetime import datetime
from shutil import copyfile
from subprocess import Popen, PIPE

import pkg_resources

from .. import resources as plotman_resources
from ..configuration import get_path

statement = 'End : {}, IO File {}'


def writeFile(content, filename):
    fo = open(filename, "w")
    fo.write(content)
    fo.close()
    print(statement.format(time.ctime(), filename))


class YamlGen:
    setting_yaml = """
    # worked for the host {host}
    # plotman version {ver}
    # last update {date}
    
    user_interface:
            use_stty_size: True
    directories:
            log: /home/ipant/.chia/mainnet
            tmp:
    {list_temp}
            {tmp2_content}
            dst:
    {list_dst}
            archive:
                    #rsyncd_module: plots
                    #rsyncd_path: /plots
                    #rsyncd_bwlimit: 80000  # Bandwidth limit in KB/s
                    #rsyncd_host: myfarmer
                    #rsyncd_user: chia
                    #index: 0
    scheduling:
            tmpdir_stagger_phase_major: 2
            tmpdir_stagger_phase_minor: 1
            tmpdir_stagger_phase_limit: 6
            tmpdir_max_jobs: 45
            global_max_jobs: 90
            global_stagger_m: [12,18]
            polling_time_s: 30
            parallel: 6

    plotting:
            k: 32
            e: True              # Use -e plotting option
            n_threads: 2         # Threads per job
            n_buckets: 128       # Number of buckets to split data into
            job_buffer: 32000     # Per job memory
            # If specified, pass through to the -f and -p options.  See CLI reference.
            farmer_pk: 861467df6768932f1c2f3a7c00a70f1e111b22b55de42c497fb93fd398eb4fd4b213f399e0930d265ffdc2b5655f96f5
            pool_pk: 8e211cb1118b95ac9da2f31fd159a74993bfcc8ab72bff79c880227185ba2b1f372b8e829bd99f5796eb4f68e41f30a9

    apis:
            api_polling_throttle_s: 5
            port: 19451
            target: "121.29.84.81"
            """

    def genTempFolder(self, d: str) -> str:
        return "                - {}/temp".format(d)

    def genDstFolder(self, d: str) -> str:
        return "                - {}/chiaFinalData".format(d)

    def temp2Hx(self, d: str) -> str:
        return "tmp2: {}".format(d)

    def __init__(self):
        self.tmp_list = []
        self.dst_list = []
        self.device_dst_list = []
        self.device_list = []

    @property
    def getDevicesDst(self) -> list:
        return self.device_dst_list

    def discoverDevices(self):
        df = Popen(["df", "-hT"], stdout=PIPE)
        output = df.communicate()[0]
        listdevice = output.split("\n")
        m = 0
        tmpnvmede = []
        nfsde = []
        for h in listdevice:
            if m > 0:
                device, type, size, used, available, percent, mountpoint = h.split()
                # if that is nvme ssd device we will take it as temp folder
                h = re.match('^\/dev\/md(\d+)', device)
                if h:
                    tmpnvmede.append(mountpoint)

                if type == "nfs4":
                    nfsde.append(mountpoint)

            m += 1

        self.device_list = tmpnvmede
        self.device_dst_list = nfsde

    def generateDummpy(self) -> None:
        if os.path.isfile(get_path()):
            overwrite = None
            while overwrite not in {"y", "n"}:
                overwrite = input(
                    f"A 'plotman.yaml' file already exists at the default location: '{get_path()}' \n\n"
                    "\tInput 'y' to overwrite existing file, or 'n' to exit without overwrite."
                ).lower()

                if overwrite == 'n':
                    print("\nExited without overrwriting file")
                    return

        # Copy the default plotman.yaml (packaged in plotman/resources/) to the user's config file path,
        # creating the parent plotman file/directory if it does not yet exist
        with importlib.resources.path(plotman_resources, "plotman.yaml") as default_config:
            config_dir = os.path.dirname(get_path())
            os.makedirs(config_dir, exist_ok=True)
            copyfile(default_config, get_path())
            print(f"\nWrote default plotman.yaml to: {get_path()}")
            return

    def ensureFolderExist(self, listp: list):
        for d in listp:
            Popen(["mkdir", "-p", d], stdout=PIPE)

    def finalized(self) -> None:
        log_file_time = datetime.strptime('Sun Apr  4 19:00:50 2021', '%a %b  %d %H:%M:%S %Y')
        version_code = pkg_resources.get_distribution('plotmanx')

        for mountpoint in self.device_list:
            self.tmp_list.append(self.genTempFolder(mountpoint))

        for mountptdst in self.device_dst_list:
            self.dst_list.append(self.genDstFolder(mountptdst))

        self.ensureFolderExist(self.tmp_list)

        self.ensureFolderExist(self.dst_list)

        k = self.setting_yaml.format(
            list_dst="\n".join(self.dst_list),
            list_temp="\n".join(self.tmp_list),
            tmp2_content=self.temp2Hx(self.tmp_list[0]),
            ver=version_code,
            date=log_file_time,
            host=socket.gethostname()
        )
        writeFile(k, get_path())
        print("==================")
        print(k)
        print("==================")



import glob
import os
import re
from subprocess import Popen, PIPE, STDOUT

from . import configuration
from .configuration import PlotmanConfig
from .util import plot_util



class FarmPlot:
    """
    The final state to move files from tmp folder to given network hard drive devices
    """

    def __init__(self, cfg: PlotmanConfig):
        self.dsts = cfg.directories.dst
        self.checktmps = cfg.directories.tmp
        self.schedule = cfg.scheduling.polling_time_s
        self.log_file_path = configuration.get_log_path()

    def start_nfs(self) -> None:

        count1 = len(glob.glob1("~/chia-blockchain", "plmo"))
        if count1 >= 1:
            for d in self.dsts:
                p = Popen(["./plmo", self.checktmps, d, 50000000, self.log_file_path], stdout=PIPE, stderr=STDOUT)
                output, error = p.communicate()
                output = output.strip().decode("utf-8")
                print("start moving file from temp folder to network FS farm location {}".format(d))
                if p.returncode:
                    print(f"E: {output}")
        else:
            print("Did not find the plmo executable.")

    def start_nvme_local_dir(self) -> None:
        subs = [x[0] for x in os.walk("/mnt")]
        count2 = len(subs)
        for rsub in subs:
            print(rsub)

    def maintainence(self, jobs: list) -> None:
        plot_util.tidy_up(jobs, self.checktmps)



import re

import psutil
from psutil.tests import sh


def check_mount_sd(mountpoint: str) -> bool:
    g = re.match("^\/dev\/sd([a-z])$", mountpoint)
    if g:
        return True
    else:
        return False


def test_disk_partitions_and_usage():
    # test psutil.disk_usage() and psutil.disk_partitions()
    # against "df -a"
    def dff(path):
        output = sh('df -P -B 1 "%s"' % path).strip()
        # df = Popen(["df", "-P", "-B", "1", path], stdout=PIPE)
        # output = df.communicate()[0]

        lines = output.split('\n')
        lines.pop(0)
        line = lines.pop(0)
        dev, total, used, free = line.split()[:4]
        if dev == 'none':
            dev = ''
        total, used, free = int(total), int(used), int(free)
        return dev, total, used, free

    for part in psutil.disk_partitions(all=False):

        dev, total, used, free = dff(part.mountpoint)
        if check_mount_sd(dev):
            usage = psutil.disk_usage(part.mountpoint)
            print(f"found - part {part.mountpoint}")
            print(dev)
            print(usage)

            # assert usage.total == total
            # 10 MB tollerance
            if abs(usage.free - free) > 10 * 1024 * 1024:
                print("psutil=%s, df=%s" % (usage.free, free))
            if abs(usage.used - used) > 10 * 1024 * 1024:
                print("psutil=%s, df=%s" % (usage.used, used))


print("before __name__ guard")
if __name__ == '__main__':
    # disks = [d.mountpoint for d in psutil.disk_partitions()]
    # disks_usage = [psutil.disk_usage(d) for d in disks]
    # print(disks_usage)

    test_disk_partitions_and_usage()

print("after __name__ guard")



import re

import psutil
from psutil.tests import sh


def commaInt(x: str) -> int:
    return int(x.strip(" ").replace(',', ''))


# if not found.. sudo: nvme: command not found
# sudo nvme smart-log /dev/nvme0
# 'NR==1{print $1}'
# lsblk | grep nvme | grep disk | sort | awk '{print $1}'
# lsblk | grep nvme | sort | awk 'NR==1{print $1}'
# lsblk|grep nvme|grep disk|sort|awk '{print $1}`
sample1 = """
nvme0n1
nvme1n1
nvme2n1
nvme3n1
"""
sample2 = """
Smart Log for NVME device:nvme0n1 namespace-id:ffffffff
critical_warning                    : 0
temperature                         : 44 C
available_spare                     : 100%
available_spare_threshold           : 10%
percentage_used                     : 6%
data_units_read                     : 1,453,614,235
data_units_written                  : 1,706,252,589
host_read_commands                  : 2,582,632,560
host_write_commands                 : 3,177,813,741
controller_busy_time                : 31,993
power_cycles                        : 7
power_on_hours                      : 3,966
unsafe_shutdowns                    : 3
media_errors                        : 0
num_err_log_entries                 : 0
Warning Temperature Time            : 0
Critical Composite Temperature Time : 0
Temperature Sensor 1                : 44 C
Temperature Sensor 2                : 44 C
Temperature Sensor 3                : 49 C
Thermal Management T1 Trans Count   : 0
Thermal Management T2 Trans Count   : 0
Thermal Management T1 Total Time    : 0
Thermal Management T2 Total Time    : 0
"""


def devNvmeList() -> list:
    devices = []
    try:
        output0 = sh("lsblk | grep nvme | grep disk | sort | awk '{print $1}'").strip()
        lines = output0.split('\n')

        for dev in lines:
            ok = re.match("^nvme(.*)", dev)
            if ok:
                print(f"sudo nvme smart-log /dev/{dev}")
                output = sh(f"sudo nvme smart-log /dev/{dev}").strip()
                if output == "sudo: nvme: command not found":
                    return []

                lines = output.split('\n')
                setting = dict()

                for l in lines:
                    if l == "":
                        continue
                    if re.match(r"^Smart Log for", l):
                        continue

                    # print("ok read. ")
                    # print(l)

                    r = l.split(":")
                    if r[0].strip(" ").strip("\n") == 'data_units_written':
                        dev_ex = r[1].strip(" ").strip("\n")
                        setting.setdefault("data_units_written", commaInt(dev_ex))

                    if r[0].strip(" ").strip("\n") == 'host_write_commands':
                        dev_ex = r[1].strip(" ").strip("\n")
                        setting.setdefault("host_write_commands", commaInt(dev_ex))

                    if r[0].strip(" ").strip("\n") == 'percentage_used':
                        dev_ex = r[1].strip(" ").strip("\n")
                        setting.setdefault("percentage_used", dev_ex.strip(" "))
                devices.append(setting)


    except UserWarning as u:
        print("not working... ps command not found")

    return devices


def crawl_exd():
    try:
        print("check --1")
        disk_counters = psutil.disk_io_counters(perdisk=True)
        for device_name in disk_counters:
            m0 = re.match("^nvme(.*)", device_name)
            if m0:
                # meta = dffnvme(device_name)
                meta = dict()
                print(f'Disk Health - {device_name}: {meta.get("host_write_commands")} b')

        print("check --3")
    except OSError as e:
        print(
            u'Caught exception when crawling disk I/O counters: {0}'.format(e))


print("before __name__ guard")
if __name__ == '__main__':
    s = devNvmeList()
    print(s)
print("after __name__ guard")



#!/usr/bin/python
#

from datetime import datetime
from os.path import isfile, join

import requests
from tornado import web, ioloop, httpserver
from tornado.template import Loader

from .servefiles import DirectoryHandler
from .sql import SQLX
from ..configuration import PlotmanConfig, get_dash_v1

__author__ = 'lousvicton'

from ..util import plot_util

try:
    import simplejson as json
except ImportError:
    import json


def APILib_data_link(dp: dict, cfg: PlotmanConfig):
    # sending post request and saving response as response object
    try:
        payload = json.dumps(dp)
        requests.post(
            url=f'http://{cfg.apis.target}:{cfg.apis.port}/api/v1/report',
            data=payload,
            timeout=5.0
        )
    except requests.Timeout as v:
        print(f"time out from{v}")
    except requests.ConnectionError as v:
        print(f"max connection error: {v}")
    except requests.HTTPError as v:
        print(f"max http unknown error: {v}")


class ApiBase(web.RequestHandler):
    SUPPORTED_METHODS = ["GET"]

    def initialize(self, *args, **kwargs):
        self.version_plotman = plot_util.plotman_version()
        self.remote_ip = self.request.headers.get('X-Forwarded-For', self.request.headers.get('X-Real-Ip', self.request.remote_ip))
        self.using_ssl = (self.request.headers.get('X-Scheme', 'http') == 'https')

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def send_response(self, data, status=200):
        """Construct and send a JSON response with appropriate status code."""
        self.set_status(status)
        t = int(datetime.now().timestamp())
        response = {'result': 1, 'data': data, 'ts': t, 'ver': self.version_plotman}
        self.write(json.dumps(response))


class NodeHandle(web.RequestHandler):
    def initialize(self, *args, **kwargs):
        self.remote_ip = self.request.headers.get('X-Forwarded-For', self.request.headers.get('X-Real-Ip', self.request.remote_ip))
        self.using_ssl = (self.request.headers.get('X-Scheme', 'http') == 'https')

    def post(self):
        self.set_header("Content-Type", "text/plain")
        req_body = self.request.body
        sq = SQLX()
        sq.schema_plan()

        jpayload = dict()

        try:
            jpayload = json.loads(req_body)
        except:
            print("error from decoding json file.")
            return

        g = datetime.now().strftime('%m-%d %H:%M:%S')
        tstime = datetime.now().timestamp()

        if req_body is not None:
            host = sq.datainput(jpayload, tstime)

            print("Report data -")
            print(f"{g} | {host} | {jpayload['version']}")

        sq.end()




class DashSimpleListNodes(ApiBase):

    def get(self):
        sq = SQLX()
        res = sq.getNodes()
        sq.end()
        # key = self.get_argument('key', None)
        self.send_response(res)


class ApiV1Review(ApiBase):

    def get(self):
        """List of routes for this API."""
        routes = {
            'info': 'GET /api/v1',
            'register': 'POST /api/v1/accounts',
            'single profile detail': 'GET /api/v1/accounts/<username>',
            'edit profile': 'PUT /api/v1/accounts/<username>',
            'delete profile': 'DELETE /api/v1/accounts/<username>',
            'login': 'POST /api/v1/accounts/login',
            'logout': 'GET /api/v1/accounts/logout',
            "user's tasks": 'GET /api/v1/accounts/<username>/tasks',
            "create task": 'POST /api/v1/accounts/<username>/tasks',
            "task detail": 'GET /api/v1/accounts/<username>/tasks/<id>',
            "task update": 'PUT /api/v1/accounts/<username>/tasks/<id>',
            "delete task": 'DELETE /api/v1/accounts/<username>/tasks/<id>'
        }
        self.write(json.dumps(routes))


class WebpHandler(web.StaticFileHandler):
    def write_error(self, status_code, **kwargs):

        if status_code in (404, 500):
            error_page = '{}.html'.format(status_code)
            if isfile(join(get_dash_v1(), error_page)):
                self.write(Loader(get_dash_v1()).load(error_page).generate())
            else:
                super().write_error(status_code, **kwargs)


# Reset
Color_Off = '\033[0m'  # Text Reset

# Regular Colors
Black = '\033[0;30m'  # Black
Red = '\033[0;31m'  # Red
Green = '\033[0;32m'  # Green
Yellow = '\033[0;33m'  # Yellow
Blue = '\033[0;34m'  # Blue
Purple = '\033[0;35m'  # Purple
Cyan = '\033[0;36m'  # Cyan
White = '\033[0;37m'  # White

# Bold
BBlack = '\033[1;30m'  # Black
BRed = '\033[1;31m'  # Red
BGreen = '\033[1;32m'  # Green
BYellow = '\033[1;33m'  # Yellow
BBlue = '\033[1;34m'  # Blue
BPurple = '\033[1;35m'  # Purple
BCyan = '\033[1;36m'  # Cyan
BWhite = '\033[1;37m'  # White

# Underline
UBlack = '\033[4;30m'  # Black
URed = '\033[4;31m'  # Red
UGreen = '\033[4;32m'  # Green
UYellow = '\033[4;33m'  # Yellow
UBlue = '\033[4;34m'  # Blue
UPurple = '\033[4;35m'  # Purple
UCyan = '\033[4;36m'  # Cyan
UWhite = '\033[4;37m'  # White

# Background
On_Black = '\033[40m'  # Black
On_Red = '\033[41m'  # Red
On_Green = '\033[42m'  # Green
On_Yellow = '\033[43m'  # Yellow
On_Blue = '\033[44m'  # Blue
On_Purple = '\033[45m'  # Purple
On_Cyan = '\033[46m'  # Cyan
On_White = '\033[47m'  # White

# High Intensity
IBlack = '\033[0;90m'  # Black
IRed = '\033[0;91m'  # Red
IGreen = '\033[0;92m'  # Green
IYellow = '\033[0;93m'  # Yellow
IBlue = '\033[0;94m'  # Blue
IPurple = '\033[0;95m'  # Purple
ICyan = '\033[0;96m'  # Cyan
IWhite = '\033[0;97m'  # White

# Bold High Intensity
BIBlack = '\033[1;90m'  # Black
BIRed = '\033[1;91m'  # Red
BIGreen = '\033[1;92m'  # Green
BIYellow = '\033[1;93m'  # Yellow
BIBlue = '\033[1;94m'  # Blue
BIPurple = '\033[1;95m'  # Purple
BICyan = '\033[1;96m'  # Cyan
BIWhite = '\033[1;97m'  # White

# High Intensity backgrounds
On_IBlack = '\033[0;100m'  # Black
On_IRed = '\033[0;101m'  # Red
On_IGreen = '\033[0;102m'  # Green
On_IYellow = '\033[0;103m'  # Yellow
On_IBlue = '\033[0;104m'  # Blue
On_IPurple = '\033[0;105m'  # Purple
On_ICyan = '\033[0;106m'  # Cyan
On_IWhite = '\033[0;107m'  # White


def start_master_api_node(cfg: PlotmanConfig):
    try:
        version = plot_util.plotman_version()
        print(f"Now plotman API - v{version}")
        print(f"API running port {cfg.apis.port} is now listening.. ")

        settings = dict(
            debug=True,
            gzip=True,
            static_path=get_dash_v1(),
        )

        """
      template_path=get_dash_v1(),
            static_path=get_dash_v1_static(),
            """
        static_handler_args = dict(
            path=get_dash_v1(),
            default_filename="index.html"
        )

        appcli = web.Application([
            (r"/api/v1/report", NodeHandle),
            (r"/api/v1/nodes", DashSimpleListNodes),
            (r"/api/v1", ApiV1Review),
            (r"/monitor/(.*)", DirectoryHandler, static_handler_args)
        ], **settings)

        print(f"Serve nuxt path {BCyan}{get_dash_v1()}{Color_Off} and set default entry point as index.html")
        print(f"Running at:{BCyan} http://{'127.0.0.1'}:{cfg.apis.port}/ {Color_Off}")
        http_server = httpserver.HTTPServer(appcli)
        http_server.listen(cfg.apis.port)
        ioloop.IOLoop.current().start()

    except OSError:
        print(f"{BRed}Port is in use.{Color_Off} Please close the daemon app and start again. ")
    except KeyboardInterrupt:
        print('Stopping server...')



import re

import psutil


def crawl_disk_io_counters():
    try:

        print("check --1")
        disk_counters = psutil.disk_io_counters(perdisk=True)
        print(disk_counters)
        for device_name in disk_counters:
            m0 = re.match("^nvme(.*)$", device_name)
            if m0:
                counters = disk_counters[device_name]
                curr_counters = [
                    counters.read_count,
                    counters.write_count,
                    counters.read_bytes,
                    counters.write_bytes
                ]

                mbpersec = counters.write_bytes / 1024 / 1024

                print(f'Disk I/O counters - {device_name}: {mbpersec} mb/s')
            # yield (device_name, curr_counters)

        print("check --3")
    except OSError as e:
        print(
            u'Caught exception when crawling disk I/O counters: {0}'.format(e))


print("before __name__ guard")
if __name__ == '__main__':
    print("check --v")
    crawl_disk_io_counters()

print("after __name__ guard")



# !/usr/bin/env python
import os

import tornado.web
from tornado.escape import xhtml_escape
# from tornado.options import define, options

# config options
"""define('port', default=8080, type=int, help='port to run web server on')
define('debug', default=True, help='start app in debug mode')
define('route_to_index', default=False, help='route all requests to index.html')
options.parse_command_line(final=True)

PORT = options.port
DEBUG = options.debug"""

ROUTE_TO_INDEX = True

class DirectoryHandler(tornado.web.StaticFileHandler):
    def validate_absolute_path(self, root, absolute_path):
        if ROUTE_TO_INDEX and self.request.uri != '/' and not '.' in self.request.uri:
            uri = self.request.uri
            if self.request.uri.endswith('/'):
                uri = uri[:-1]

            absolute_path = absolute_path.replace(uri, '/index.html')

        if os.path.isdir(absolute_path):
            index = os.path.join(absolute_path, 'index.html')
            if os.path.isfile(index):
                return index

            return absolute_path

        return super(DirectoryHandler, self).validate_absolute_path(root, absolute_path)

    def get_content_type(self):
        if self.absolute_path.endswith('.vtt'):
            return 'text/vtt'

        if self.absolute_path.endswith('.m3u8'):
            return 'application/vnd.apple.mpegurl'

        content_type = super(DirectoryHandler, self).get_content_type()

        # default to text/html
        if content_type == 'application/octet-stream':
            return 'text/html'

        return content_type

    @classmethod
    def get_content(cls, abspath, start=None, end=None):
        relative_path = abspath.replace(os.getcwd(), '') + '/'

        if os.path.isdir(abspath):
            html = '<html><title>Directory listing for %s</title><body><h2>Directory listing for %s</h2><hr><ul>' % (relative_path, relative_path)
            for filename in os.listdir(abspath):
                force_slash = ''
                full_path = filename
                if os.path.isdir(os.path.join(relative_path, filename)[1:]):
                    full_path = os.path.join(relative_path, filename)
                    force_slash = '/'

                html += '<li><a href="%s%s">%s%s</a>' % (xhtml_escape(full_path), force_slash, xhtml_escape(filename), force_slash)

            return html + '</ul><hr>'

        if os.path.splitext(abspath)[1] == '.md':
            try:
                import codecs
                import markdown
                input_file = codecs.open(abspath, mode='r', encoding='utf-8')
                text = input_file.read()
                return markdown.markdown(text)
            except:
                import traceback
                traceback.print_exc()
                pass

        return super(DirectoryHandler, cls).get_content(abspath, start=start, end=end)



import re
from subprocess import Popen, PIPE

print("before __name__ guard")
if __name__ == '__main__':

    df = Popen(["df", "-hT"], stdout=PIPE)
    output = df.communicate()[0]
    listdevice = output.split("\n")
    m = 0
    tmpnvmede = []
    nfsde = []
    for h in listdevice:
        if m > 0:
            device, type, size, used, available, percent, mountpoint = h.split()
            # if that is nvme ssd device we will take it as temp folder
            h = re.match('^\/dev\/md(\d+)', device)
            if h:
                tmpnvmede.append(mountpoint)

            if type == "nfs4":
                nfsde.append(mountpoint)

        m += 1

    print(nfsde)
print("after __name__ guard")



from source.plotmanx import configuration
from source.plotmanx.query.api import start_master_api_node

print("before __name__ guard")
if __name__ == '__main__':
    cfg = configuration.get_validated_configs()
    start_master_api_node(cfg)
print("after __name__ guard")



import contextlib
import math
import os
import re

import pendulum
import pkg_resources
import psutil
from psutil.tests import sh

GB = 1_000_000_000

# IN GB size
KSIZE = {
    "32": 256.6,
    "33": 550,
    "34": 1118,
    "35": 2335,
}

# IN GB size
PLOT_SIZE = {
    "32": 108.837,
    "33": 224.227,
    "34": 461.535,
    "35": 949.3,
}

PLOT_LIST = {
    "32": r'^plot-k32-.*plot$',
    "33": r'^plot-k33-.*plot$',
    "34": r'^plot-k34-.*plot$',
    "35": r'^plot-k35-.*plot$',
}

PROGRESS_BAR = {
    "phase1_line_end": 801,
    "phase2_line_end": 834,
    "phase3_line_end": 2474,
    "phase4_line_end": 2620,
    "phase1_weight": 33.4,
    "phase2_weight": 20.43,
    "phase3_weight": 42.29,
    "phase4_weight": 3.88,
}


def is_phase_start(phase_pair) -> bool:
    (ph, subph) = phase_pair
    return ph < 3 and subph < 6


def phase_str(phase_pair) -> str:
    (ph, subph) = phase_pair
    return ((str(ph) if ph is not None else '?') + ':'
            + (str(subph) if subph is not None else '?'))


def abbr_path(path, putative_prefix) -> str:
    if putative_prefix and path.startswith(putative_prefix):
        return os.path.relpath(path, putative_prefix)
    else:
        return path


def df_b(d) -> int:
    """
    Return free space for directory (in bytes)
    """
    stat = os.statvfs(d)
    return stat.f_frsize * stat.f_bavail


def checkTempSpace(k_number: int, tmpdir: str) -> bool:
    for k, v in KSIZE.items():
        if int(k) == k_number:
            gb_free = df_b(tmpdir) / GB
            if gb_free > float(v):
                return True
    return False


def isSpaceCritical(tmpdir: str) -> bool:
    """
    determine when the current working directory is on critical free space limitzxzx
    """

    gb_free = df_b(tmpdir) / GB
    if gb_free < 1:
        return True
    else:
        return False


def availableSpace(tmpdir: str) -> str:
    gb_free = df_b(tmpdir) / GB
    return f"{gb_free} GB"


def get_k32_plotsize() -> float:
    v = PLOT_SIZE["32"]
    return v * GB


def human_format(num, precision):
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    return (('%.' + str(precision) + 'f%s') %
            (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude]))


def time_format(sec):
    if sec is None:
        return '-'
    if sec < 60:
        return '%ds' % sec
    else:
        return '%d:%02d' % (int(sec / 3600), int((sec % 3600) / 60))


def tmpdir_phases_str(tmpdir_phases_pair):
    tmpdir = tmpdir_phases_pair[0]
    phases = tmpdir_phases_pair[1]
    phase_str = ', '.join(['%d:%d' % ph_subph for ph_subph in sorted(phases)])
    return ('%s (%s)' % (tmpdir, phase_str))


def getMaxFitablePlots(drive_size: int, final_plot_size: int) -> float:
    driveSpaceBuffer = 0.05
    amt = ((drive_size - (drive_size * driveSpaceBuffer))) / final_plot_size
    return amt


def availableSizeDriveForPlots(sized: int, final_plot_size: int) -> float:
    return sized / final_plot_size


def split_path_prefix(items):
    if not items:
        return ('', [])

    prefix = os.path.commonpath(items)
    if prefix == '/':
        return ('', items)
    else:
        remainders = [os.path.relpath(i, prefix) for i in items]
        return (prefix, remainders)


def list_k32_plots(d) -> list:
    """List completed k32 plots in a directory (not recursive)"""
    return list_k_plots(d, 32)


def list_k_plots(d: str, K: int) -> list:
    plots = []
    for plot in os.listdir(d):
        if re.match(PLOT_LIST[str(K)], plot):
            plot = os.path.join(d, plot)
            try:
                require_size = PLOT_SIZE[str(K)] * 0.95 * GB
                if os.stat(plot).st_size > require_size:
                    plots.append(plot)
            except FileNotFoundError:
                continue

    return plots


def column_wrap(items, n_cols, filler=None) -> list:
    '''Take items, distribute among n_cols columns, and return a set
       of rows containing the slices of those columns.'''
    rows = []
    n_rows = math.ceil(len(items) / n_cols)
    for row in range(n_rows):
        row_items = items[row:: n_rows]
        # Pad and truncate
        rows.append((row_items + ([filler] * n_cols))[:n_cols])
    return rows


def get_frw_status(job) -> str:
    f = "Y" if job.isFrozen else "N"
    r = "Y" if job.isReadFail else "N"
    w = "Y" if job.isWroteErr else "N"
    return f"{f}/{r}/{w}"


def parse_chia_plot_time(s):
    # This will grow to try ISO8601 as well for when Chia logs that way
    return pendulum.from_format(s, 'ddd MMM DD HH:mm:ss YYYY', locale='en', tz=None)


def countnonoverlappingrematches(pattern, thestring):
    return re.subn(pattern, '', thestring)[1]


def chia_version() -> str:
    plot_args = ['chia', 'version']
    try:
        version = os.popen(" ".join(plot_args), 'r').read().rstrip("\n")
    except TypeError:
        version = 'undetected'
    except ValueError:
        version = 'undetected'

    return version


def plotman_version() -> str:
    return pkg_resources.get_distribution('plotmanx').version


def get_plot_progress(line_count: int) -> float:
    phase1_line_end = PROGRESS_BAR.get('phase1_line_end')
    phase2_line_end = PROGRESS_BAR.get('phase2_line_end')
    phase3_line_end = PROGRESS_BAR.get('phase3_line_end')
    phase4_line_end = PROGRESS_BAR.get('phase4_line_end')
    phase1_weight = PROGRESS_BAR.get('phase1_weight')
    phase2_weight = PROGRESS_BAR.get('phase2_weight')
    phase3_weight = PROGRESS_BAR.get('phase3_weight')
    phase4_weight = PROGRESS_BAR.get('phase4_weight')
    progress = 0
    if line_count > phase1_line_end:
        progress += phase1_weight
    else:
        progress += phase1_weight * (line_count / phase1_line_end)
        return progress
    if line_count > phase2_line_end:
        progress += phase2_weight
    else:
        progress += phase2_weight * ((line_count - phase1_line_end) / (phase2_line_end - phase1_line_end))
        return progress
    if line_count > phase3_line_end:
        progress += phase3_weight
    else:
        progress += phase3_weight * ((line_count - phase2_line_end) / (phase3_line_end - phase2_line_end))
        return progress
    if line_count > phase4_line_end:
        progress += phase4_weight
    else:
        progress += phase4_weight * ((line_count - phase3_line_end) / (phase4_line_end - phase3_line_end))

    return progress


def tidy_up(jobs: list, temp_folders: list) -> None:
    active_plot_ids = [r.plot_id for r in jobs]
    count_files = 0
    remove_paths = []

    for d in temp_folders:
        print(f"🚧 check {d}")
        with os.scandir(d) as it:
            for entry in it:
                if not entry.name.endswith('.plot') and entry.is_file() and not entry.name.endswith(".db"):
                    print(f"📥 check file {entry.name}")
                    if len(active_plot_ids) > 0:
                        found = False
                        for actId in active_plot_ids:
                            if actId in entry.name:
                                found = True
                                break

                        if not found:
                            print(f"✅ qualified file {entry.name}")
                            remove_paths.append(entry.path)
                    else:
                        print(f"✅ qualified file {entry.name}")
                        remove_paths.append(entry.path)

    print("-------------------------")
    print("Active plot ids:")
    print(active_plot_ids)

    for u in remove_paths:
        try:
            os.unlink(u)
            count_files = count_files + 1
            print(f"Found and removed unrelated tmp {u}...")
        except FileNotFoundError:
            print(f"Failed to remove file {u}")
            pass

    print("-------------------------")
    print(f"complete total {count_files} files removed")


def discover_local_hhd() -> list:
    # test psutil.disk_usage() and psutil.disk_partitions()
    # against "df -a"
    def dff(path):
        output = sh('df -P -B 1 "%s"' % path).strip()
        # df = Popen(["df", "-P", "-B", "1", path], stdout=PIPE)
        # output = df.communicate()[0]

        lines = output.split('\n')
        lines.pop(0)
        line = lines.pop(0)
        dev, total, used, free = line.split()[:4]
        if dev == 'none':
            dev = ''
        total, used, free = int(total), int(used), int(free)
        return dev, total, used, free

    devices = []
    for part in psutil.disk_partitions(all=False):
        dev, total, used, free = dff(part.mountpoint)

        g = re.match("^\/dev\/sd([a-z])$", dev)

        if g:
            usage = psutil.disk_usage(part.mountpoint)
            print(f"found - {dev} {part.mountpoint} {usage}")
            percent = float(used / total * 100)

            # assert usage.total == total
            # 10 MB tollerance

            if abs(usage.free - free) > 10 * 1024 * 1024:
                print("psutil=%s, df=%s" % (usage.free, free))
            if abs(usage.used - used) > 10 * 1024 * 1024:
                print("psutil=%s, df=%s" % (usage.used, used))

            lowCap = False
            if usage.free < 100 * GB:
                lowCap = True

            devices.append(dict(disk=dev, per=percent, low=lowCap))

    return devices


def discover_nvme_io() -> list:
    devices = []
    try:
        disk_counters = psutil.disk_io_counters(perdisk=True)
        # print(disk_counters)
        for device_name in disk_counters:
            m0 = re.match("^nvme(.*)$", device_name)
            if m0:
                counters = disk_counters[device_name]
                gbpersec = float(counters.write_bytes) / GB
                print(f'Disk I/O counters device - {device_name}: {gbpersec} gb/s')
                devices.append(dict(dev=device_name, speed=gbpersec))

    except OSError as e:
        print(u'Caught exception when crawling disk I/O counters: {0}'.format(e))
    except TypeError as v:
        print("there is an type error.")

    return devices


def is_plot_moving(cmdline: list) -> bool:
    return (
            len(cmdline) > 3
            and cmdline[0].endswith('plmo')
    )


def is_plmo_nfs(cmdline: list) -> bool:
    test = " ".join(cmdline)
    m = re.match(r"^\/mnt\/nfs.*.?(\d+).\/", test)
    if m:
        return True
    else:
        return False


def discover_plmo_operations():
    jobs = []

    for proc in psutil.process_iter(['pid', 'cmdline']):
        # Ignore processes which most likely have terminated between the time of iteration and data access.
        with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
            if is_plot_moving(proc.cmdline()):
                job = " ".join(proc.cmdline())
                jobs.append(job)

    return jobs


def discover_nfs_operations() -> list:
    nfs_list = []
    try:
        for proc in psutil.process_iter(['pid', 'cmdline']):
            # Ignore processes which most likely have terminated between the time of iteration and data access.
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied):
                if is_plot_moving(proc.cmdline()):
                    if is_plmo_nfs(proc.cmdline()):
                        test = " ".join(proc.cmdline())
                        print(f"check path: {test}")
                        g = re.match('(\d+)', test)
                        nfs_list.append(f"192.168.10.{g[0]}")

    except TypeError as b:
        print(f"Type Error: {b}")
    return nfs_list



import psutil

from source.plotmanx.util.yamlgen import YamlGen

print("before __name__ guard")
if __name__ == '__main__':

    y = YamlGen()
    y.discoverDevices()
    print(y.getDevicesDst)

    disks = [d.mountpoint for d in psutil.disk_partitions()]
    system = {
        'boot_time': psutil.boot_time(),
        'cpu_count': psutil.cpu_count(),
        # 'cpu_stats': psutil.cpu_stats().__dict__,
        # 'cpu_times': [k.__dict__ for k in psutil.cpu_times(percpu=True)],
        # 'disk_io_counters': psutil.disk_io_counters().__dict__,
        'disk_usage': [],
        # 'net_io_counters': psutil.net_io_counters().__dict__,
        # 'swap_memory': psutil.swap_memory().__dict__,
        # 'virtual_memory': psutil.virtual_memory().__dict__
    }

    # disks_usage = [psutil.disk_usage(d) for d in disks]
    INCLUDED_PARTITIONS = set(['/', '/writable'])

    partitions = psutil.disk_partitions()
    for p in partitions:
        if p.mountpoint in INCLUDED_PARTITIONS:
            usage = psutil.disk_usage(p.mountpoint)
            system['disk_usage'].append({
                'mountpoint': p.mountpoint,
                'total': usage.total,
                'used': usage.used
            })

    print(system)
print("after __name__ guard")



import json
import sqlite3
from datetime import datetime

from ..configuration import get_db_path


def txtBlock(s: any) -> str:
    if s:
        return str(s)
    else:
        return ""


def txtJson(s: any) -> str:
    return json.dumps(s)


def commaInt(x: str) -> int:
    return int(x.replace(',', ''))


class SQLX:
    def __init__(self):
        self.con = sqlite3.connect(get_db_path())
        self.cur = self.con.cursor()

    def end(self):
        self.con.commit()
        self.con.close()

    def schema_plan(self):

        self.cur.execute('''CREATE TABLE IF NOT EXISTS nodeInfo (
        
                                   
                                   tid INTEGER PRIMARY KEY AUTOINCREMENT,
                                   host text NOT NULL,
                                   
                                   
                                   tmp_plots INTEGER NOT NULL,
                                   mv_channels INTEGER NOT NULL,
                                   cpu_count INTEGER NOT NULL,
                                   cpu_percent REAL NOT NULL,
                                   bucket_sum_t REAL NOT NULL,
                                   bucket_count INTEGER NOT NULL,
                                   
                                   swap_percent REAL NOT NULL,
                                   disk_percent_block text NOT NULL,
                                   io_nvme_block text NOT NULL,
                                   io_list_plmo text NOT NULL,
                                   io_list_nfs text NOT NULL,
                                   io_read_issues INTEGER NOT NULL,
                                   
                                   memory_percent REAL NOT NULL,
                                                                      
                                   net_read_gb_s REAL NOT NULL,
                                   net_write_gb_s REAL NOT NULL,
                                   disk_read_gb_s REAL NOT NULL,
                                   disk_write_gb_s REAL NOT NULL,

                                   net_fds INTEGER NOT NULL,
                                   version text NOT NULL,
                                   chia_ver text NOT NULL,
                                   timen INTEGER NOT NULL
                                   
                                
                           );''')

        self.cur.execute('''CREATE TABLE IF NOT EXISTS plotv2 (
                           pcid INTEGER PRIMARY KEY AUTOINCREMENT,
                           plotid text NOT NULL,
                           k INTEGER NOT NULL,
                           r INTEGER NOT NULL,
                           b INTEGER NOT NULL,
                           u INTEGER NOT NULL,
                           pid INTEGER NOT NULL,
                           progress REAL NOT NULL,
                           phase text NOT NULL,
                           host text NOT NULL,
                           freeze text NOT NULL,
                           readerr text NOT NULL,
                           wroteerr text NOT NULL,
                           time INTEGER NOT NULL
                   );''')

    """
    host name as return
    """

    def datainput(self, j: dict, ts: int) -> str:
        host = txtBlock(j['identity'])
        if len(j['jobls']) > 0:

            try:
                for h in j['jobls']:

                    plot_id = h['plot_id']

                    content_find = f"""

                    SELECT COUNT(*) FROM plotv2 WHERE plotid='{plot_id}';

                    """

                    (n,) = self.cur.execute(content_find).fetchone()

                    if int(n) == 0:
                        content_insert = f"""
                       INSERT INTO plotv2 (
                            plotid, k, r, b, u, pid, 
                            progress, phase,
                            freeze,readerr,wroteerr,
                            host, time)
                            VALUES (
                            '{plot_id}', {int(h['k'])}, {int(h['r'])}, {int(h['b'])}, {int(h['u'])}, {int(h['pid'])}, 
                            {float(h['progress'])}, '{txtBlock(h['phase'])}',
                            '{txtBlock(h['freeze'])}','{txtBlock(h['readerr'])}','{txtBlock(h['wroteerr'])}',
                            '{host}', {ts}
                            )
                           ;
                        """

                        self.cur.execute(content_insert)

                        print(f"ID - {plot_id}")
                    else:
                        content_update = f"""
                        UPDATE plotv2

                            SET time={ts},
                            progress={float(h['progress'])},
                            phase='{txtBlock(h['phase'])}',
                            freeze='{txtBlock(h['freeze'])}',
                            readerr='{txtBlock(h['readerr'])}',
                            wroteerr='{txtBlock(h['wroteerr'])}'

                        WHERE plotid='{plot_id}'
                        ;
                        """
                        self.cur.execute(content_update)

            except sqlite3.OperationalError as r:
                print(f"there is a things that doesnt work on - job. {r}")
        else:
            print("body is not empty")

        try:

            content_insert = f"""
            INSERT INTO nodeInfo (

                host,tmp_plots,mv_channels,cpu_count,cpu_percent,bucket_sum_t,bucket_count,swap_percent,

                disk_percent_block,io_nvme_block,io_list_plmo,io_list_nfs,io_read_issues,

                memory_percent,net_read_gb_s,net_write_gb_s,disk_read_gb_s,disk_write_gb_s,

                version,chia_ver,

                net_fds,timen

            ) VALUES(

                '{host}',
                {int(j['plotcount'])},
                {int(j['movingcount'])},
                {int(j['cpucount'])},
                {float(j['cpu_percent'])},
                {float(j['sizet'])},
                {float(j['historyplots'])},
                {float(j['swap_percent'])},

                '{txtJson(j['disk_info'])}',
                '{txtJson(j['disk_nvme_io'])}',
                '{txtJson(j['movingdetail'])}',
                '{txtJson(j['nfsips'])}',
                {int(j['io_read_issues'])},

                {float(j['memory_percent'])},
                {float(j['net_read_mb_s'])},
                {float(j['net_write_mb_s'])},
                {float(j['disk_read_mb_s'])},
                {float(j['disk_write_mb_s'])},
                
                '{txtBlock(j['version'])}',
                '{txtBlock(j['chia_ver'])}',
                0,
                {int(datetime.now().timestamp())}
            )
            """

            self.cur.execute(content_insert)

        except sqlite3.OperationalError as r:
            print(f"there is a things that doesnt work - machine. {r}")

        return host

    def getNodes(self) -> list:
        block = """
        SELECT * FROM nodeInfo ORDER BY timen DESC GROUP BY host
        """

        listd = self.cur.execute(block).fetchall()

        """for company, orders_iter in itertools.groupby(listd, key=lambda r: r[0]):
            orders = list(orders_iter)
            total_qty = sum(order[2] for order in orders)"""

        return listd



from source.plotmanx.analyze.pri import LogFile

print("before __name__ guard")
if __name__ == '__main__':
    l = LogFile("test_Data_4.log")
    l.init_logfile()
    print("{:.3g}%".format(l.getProgressPercentage))
    print("read failure:")
    print(l.failure_read)
    print("write failure")
    print(l.disk_confirm)
    print("phase")
    print(l.getPhase)
print("after __name__ guard")



import argparse
import concurrent.futures
import os
import random
import time

# Plotman libraries
from marshmallow import ValidationError

from . import analyze, archive, configuration, interactive, manager, reporting
from .analyze.report import analyze
from .clockwork import MintJ
from .configuration import PlotmanConfig
from .farmplot import FarmPlot
from .job import Job
from .query.api import start_master_api_node, APILib_data_link
from .util import plot_util
from .util.yamlgen import YamlGen


class PlotmanArgParser:
    def add_idprefix_arg(self, subparser):
        subparser.add_argument(
            'idprefix',
            type=str,
            nargs='+',
            help='disambiguating prefix of plot ID')

    def parse_args(self):
        parser = argparse.ArgumentParser(description='Chia plotting manager.')
        sp = parser.add_subparsers(dest='cmd')

        sp.add_parser('version', help='print the version')
        sp.add_parser('status', help='show current plotting status')
        sp.add_parser('dirs', help='show directories info')
        sp.add_parser('interactive', help='run interactive control/monitoring mode')
        sp.add_parser('dsched', help='print destination dir schedule')
        sp.add_parser('plot', help='run plotting loop')
        sp.add_parser('tidy', help='manually clean up the unrelated tmp files')
        sp.add_parser('api', help='open api port for external applications')
        sp.add_parser('nfs', help='using go version temp file archive')
        sp.add_parser('archive', help='move completed plots to farming location')
        p_config = sp.add_parser('config', help='display or generate plotman.yaml configuration')
        sp_config = p_config.add_subparsers(dest='config_subcommand')
        sp_config.add_parser('generate', help='generate a default plotman.yaml file and print path')
        sp_config.add_parser('path', help='show path to current plotman.yaml file')

        p_details = sp.add_parser('details', help='show details for job')
        self.add_idprefix_arg(p_details)

        p_files = sp.add_parser('files', help='show temp files associated with job')
        self.add_idprefix_arg(p_files)

        p_kill = sp.add_parser('kill', help='kill job (and cleanup temp files)')
        self.add_idprefix_arg(p_kill)

        p_suspend = sp.add_parser('suspend', help='suspend job')
        self.add_idprefix_arg(p_suspend)

        p_resume = sp.add_parser('resume', help='resume suspended job')
        self.add_idprefix_arg(p_resume)

        p_analyze = sp.add_parser('analyze', help='analyze timing stats of completed jobs')

        p_analyze.add_argument('--clipterminals',
                               action='store_true',
                               help='Ignore first and last plot in a logfile, useful for '
                                    'focusing on the steady-state in a staggered parallel '
                                    'plotting test (requires plotting  with -n>2)')
        p_analyze.add_argument('--bytmp',
                               action='store_true',
                               help='slice by tmp dirs')
        p_analyze.add_argument('--bybitfield',
                               action='store_true',
                               help='slice by bitfield/non-bitfield sorting')
        p_analyze.add_argument('logfile', type=str, nargs='+',
                               help='logfile(s) to analyze')

        args = parser.parse_args()
        return args


def get_term_width():
    columns = 0
    try:
        (rows, columns) = os.popen('stty size', 'r').read().split()
        columns = int(columns)
    except:
        columns = 120  # 80 is typically too narrow.  TODO: make a command line arg.
    return columns


def plotting(cfg: PlotmanConfig):
    print('starting plot loop')
    minp = MintJ(cfg.plotting)
    minp.PlotDaemon()
    minp.GenJobs(cfg.directories)
    minp.Upcfg(cfg.scheduling, cfg.plotting)
    minp.SpaceManagement()
    while True:
        try:
            print(f"s{cfg.scheduling.polling_time_s}\n")
            time.sleep(cfg.scheduling.polling_time_s)

            minp.GenJobs(cfg.directories)
            minp.Upcfg(cfg.scheduling, cfg.plotting)

            if minp.isCreateNewJobParallelReady:
                minp.ParallelWorker()

            if cfg.apis.target is not "":
                APILib_data_link(minp.NeuoInfo(
                    minp.LsJobs,
                    plot_util.discover_local_hhd(),
                    plot_util.discover_nvme_io()
                ), cfg)

            minp.SpaceManagement()

        except ValidationError as te:
            print('got ConfigurationException from io', te)
            continue
        except ConnectionError as cc:
            print('got ConnectionError from io', cc)
            continue
        except ValueError as v:
            print('got ValueError from io', v)
            continue
        except TimeoutError as th:
            print('got TimeoutError from io', th)
            continue
        except IOError as io:
            print('got IOError from io', io)
            continue

    print('exit from error unknown...')


def archivePlots(cfg: any):
    print('...starting archive loop')
    firstit = True
    jobs = Job.genTasks(cfg.directories.log)
    while True:
        if not firstit:
            print('Sleeping 60s until next iteration...')
            time.sleep(60)
            jobs = Job.genTasks(cfg.directories.log)
        firstit = False
        archiving_status, log_message = archive.spawn_archive_process_rsync(cfg.directories, jobs)
        if log_message:
            print('%s, %s' % (archiving_status, log_message))


def main():
    random.seed()

    pm_parser = PlotmanArgParser()
    args = pm_parser.parse_args()

    if args.cmd == 'version':
        print(plot_util.plotman_version())
        return

    elif args.cmd == 'config':
        config_file_path = configuration.get_path()
        if args.config_subcommand == 'path':
            if os.path.isfile(config_file_path):
                print(config_file_path)
                return
            print(f"No 'plotman.yaml' file exists at expected location: '{config_file_path}'")
            print(f"To generate a default config file, run: 'plotman config generate'")
            return 1
        if args.config_subcommand == 'generate':
            if os.path.isfile(config_file_path):
                overwrite = None
                while overwrite not in {"y", "n"}:
                    overwrite = input(
                        f"A 'plotman.yaml' file already exists at the default location: '{config_file_path}' \n\n"
                        "\tInput 'y' to overwrite existing file, or 'n' to exit without overwrite."
                    ).lower()
                    if overwrite == 'n':
                        print("\nExited without overrwriting file")
                        return

            yamlGenerator = YamlGen()
            yamlGenerator.discoverDevices()
            yamlGenerator.finalized()

        if not args.config_subcommand:
            print("No action requested, add 'generate' or 'path'.")
            return

    cfg = configuration.get_validated_configs()
    #
    # Stay alive, spawning plot jobs
    #
    if args.cmd == 'plot':
        plotting(cfg)

    #
    # Analysis of completed jobs
    #
    elif args.cmd == 'analyze':
        analyze(args.logfile, args.clipterminals, args.bytmp, args.bybitfield)

    else:
        jobs = Job.genTasks(cfg.directories.log)
        farm = FarmPlot(cfg)

        # Status report
        if args.cmd == 'status':
            print(reporting.status_report(jobs, get_term_width()))

        # Directories report
        elif args.cmd == 'dirs':
            print(reporting.dirs_report(jobs, cfg.directories, cfg.scheduling, get_term_width()))

        elif args.cmd == 'interactive':
            interactive.run_interactive()

        elif args.cmd == 'api':
            start_master_api_node(cfg)

        elif args.cmd == 'nfs':
            farm.start_nfs()

        elif args.cmd == 'tidy':
            farm.maintainence(jobs)

        # Start running archival
        elif args.cmd == 'archive':
            with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
                executor.submit(archivePlots, cfg)

        # Debugging: show the destination drive usage schedule
        elif args.cmd == 'dsched':
            for (d, ph) in manager.dstdirs_to_furthest_phase(jobs).items():
                print('  %s : %s' % (d, str(ph)))

        #
        # Job control commands
        #
        elif args.cmd in ['details', 'files', 'kill', 'suspend', 'resume']:
            print(args)

            selected = []

            # TODO: clean up treatment of wildcard
            if args.idprefix[0] == 'all':
                selected = jobs
            else:
                # TODO: allow multiple idprefixes, not just take the first
                selected = manager.select_jobs_by_partial_id(jobs, args.idprefix[0])
                id_spec = args.idprefix[0]
                if len(selected) == 0:
                    print('Error: %s matched no jobs.' % id_spec)
                elif len(selected) > 1:
                    print('Error: "%s" matched multiple jobs:' % id_spec)
                    for j in selected:
                        print('  %s' % j.plot_id)
                    selected = []

            for job in selected:
                if args.cmd == 'details':
                    print(job.status_str_long())

                elif args.cmd == 'files':
                    temp_files = job.get_temp_files()
                    for f in temp_files:
                        print('  %s' % f)

                elif args.cmd == 'kill':
                    # First suspend so job doesn't create new files
                    print('Pausing PID %d, plot id %s' % (job.proc.pid, job.plot_id))
                    job.suspend()

                    temp_files = job.get_temp_files()
                    print('Will kill pid %d, plot id %s' % (job.proc.pid, job.plot_id))
                    print('Will delete %d temp files' % len(temp_files))
                    conf = input('Are you sure? ("y" to confirm): ')
                    if conf != 'y':
                        print('canceled.  If you wish to resume the job, do so manually.')
                    else:
                        print('killing...')
                        job.cancel()
                        print('cleaing up temp files...')
                        for f in temp_files:
                            os.remove(f)

                elif args.cmd == 'suspend':
                    print('Suspending ' + job.plot_id)
                    job.suspend()

                elif args.cmd == 'resume':
                    print('Resuming ' + job.plot_id)
                    job.resume()
